const HyperInterface = require('hyperpubee-hyper-interface')
const { DiGraph } = require('.')
const ram = require('random-access-memory')
const Corestore = require('corestore')

const corestore = new Corestore(ram)
const hyperInterface = new HyperInterface(corestore)

async function runExample () {
  // Creating a new DiGraph with only an empty root node
  const diGraph = await DiGraph.createNew({
    name: 'My digraph',
    hyperInterface
  })

  // Set the content of a node
  await diGraph.rootNode.setContent('I am the root node')

  // A DiGraph can also be loaded from a compatible hyperbee
  const bee = await hyperInterface.createHyperbee('My other digraph')
  await bee.put(
    'root',
    { content: 'I am the root of the other digraph', children: [{ location: 'aChild' }] }
  )
  await bee.put(
    'aChild',
    { content: 'I am a child of the other digraph', children: [] }
  )
  const otherDiGraph = await DiGraph.loadFromBee(
    { hash: bee.feed.key, hyperInterface }
  )

  // Checks that the bee is indeed a valid digraph
  // (expensive operation: iterates over all children)
  await otherDiGraph.deepValidate()

  console.log(await otherDiGraph.rootNode.getChildren()) // 1 entry at location 'aChild'

  // Add a new local node as child
  const localChild = await diGraph.rootNode.createChild('localChild')
  await localChild.setContent('I am a local child')

  // Add a new remote node to a child
  // (Note: the digraph is extended with this node and all its children)
  await diGraph.rootNode.pushChild({ hash: bee.feed.key, location: 'root' })

  // Walk the di-graph (note: no order guarantees)
  for await (const node of diGraph.walkNoRepeatVisits()) {
    console.log(node.entryReq.toStr(), ': ', (await node.getContent()))
  }
  // Prints something like
  /*
    e363df1564efcbdc4fb70de864f468fbc7414e50b22904235e2b709e879c2555/root :  I am the root node
    e363df1564efcbdc4fb70de864f468fbc7414e50b22904235e2b709e879c2555/localChild :  I am a local child
    8126da769316af7073182b65d6f1f5020da0f40760a2466d5439799eb057cac3/root :  I am the root of the other digraph
    8126da769316af7073182b65d6f1f5020da0f40760a2466d5439799eb057cac3/aChild :  I am a child of the other digraph
  */
}

runExample()
