const EntryReq = require('./lib/entry-request')
const Node = require('./lib/node')
const c = require('./lib/constants')
const BeeInterface = require('./lib/bee-interface')

class DiGraph {
  constructor ({ rootEntryReq, backend, NodeClass = Node }) {
    this.rootEntryReq = new EntryReq({ ...rootEntryReq })
    this.backend = backend
    this.NodeClass = NodeClass
  }

  get rootHash () {
    return this.rootEntryReq.hash
  }

  get rootNode () {
    return new this.NodeClass({ entryReq: this.rootEntryReq, backend: this.backend })
  }

  async deepValidate () {
    const nodeEntryPromises = []
    for await (const node of this.walkNoRepeatVisits()) {
      nodeEntryPromises.push(node.getEntry())
    }

    // To check the leaf-nodes have valid entries
    await Promise.all(nodeEntryPromises)
  }

  async * walkNoRepeatVisits () {
    // Walks the di-graph starting from the root, and yields
    // each node once, without guaranteeing any order.
    // If a node is reachable from multiple paths,
    // all paths will reach that node, but only the first
    // one to reach it will yield it and continue with its children
    // --children are resolved 'simultaneously' in async sense
    // (the algorithm does not descend child per child, but requests all
    // and yields nodes in the order in which they are resolved)
    yield * this._walkRecHelper(
      { visitedEntries: new Set(), entryReq: this.rootEntryReq }
    )
  }

  async * _walkRecHelper ({ visitedEntries, entryReq }) {
    // DEVNOTE: all recursive invocations share
    // the same visitedEntries set. There must not be any
    // await between the check on whether it contains an entry,
    // and adding that entry to it. Otherwise entries might be visited
    // multiple times.
    if (visitedEntries.has(entryReq.uniqueId)) return
    visitedEntries.add(entryReq.uniqueId)

    const node = new this.NodeClass({ entryReq, backend: this.backend })
    yield node

    const childGenerators = []
    const allPromises = {}
    for (const child of (await node.getChildren())) {
      const childGenerator = this._walkRecHelper({ visitedEntries, entryReq: child })
      childGenerators.push(childGenerator)

      // DEVNOTE: i is wrapped in promise as well, to know upon resolution
      // what the corresponding childGenerator index was
      const i = childGenerators.length - 1
      allPromises[i] = Promise.all([i, childGenerator.next()])
    }

    while (Object.keys(allPromises).length > 0) {
      const [winnerIndex, winner] = await Promise.race(Object.values(allPromises))

      if (winner.done === true) {
        delete allPromises[winnerIndex]
      } else {
        allPromises[winnerIndex] = Promise.all(
          [winnerIndex, childGenerators[winnerIndex].next()]
        )
        yield winner.value
      }
    }
  }

  static async createNew ({ name, hyperInterface, ...opts }) {
    const bee = await hyperInterface.createHyperbee(name)
    await bee.put(c.ROOT, Node.getEmptyNode())

    const rootEntryReq = _getRootEntryReq(bee.feed.key)
    const backend = new BeeInterface(hyperInterface)

    return new DiGraph({ backend, rootEntryReq, ...opts })
  }

  static async loadFromBee ({ hash, hyperInterface, ...opts }) {
    const backend = new BeeInterface(hyperInterface)

    const rootEntryReq = _getRootEntryReq(hash)
    await rootEntryReq.get(backend.getEntry) // Ensures the root is valid

    return new DiGraph({ backend, rootEntryReq, ...opts })
  }
}

function _getRootEntryReq (hash) {
  return new EntryReq(
    { hash, location: c.ROOT }
  )
}

module.exports = { DiGraph, Node, BeeInterface, EntryReq }
