const { expect } = require('chai')

const { hyperInterfaceFactory } = require('./fixtures')
const EntryReq = require('../lib/entry-request')
const utils = require('../lib/utils')
const c = require('../lib/constants')

describe('Test entry-req', function () {
  const location = 'loc'
  const hash = 'a'.repeat(64)

  describe('Returns the expected uniqueId', function () {
    it('Returns the correct unique id if no version specified', function () {
      const e = new EntryReq({ hash, location })
      expect(e.uniqueId).to.equal(`${hash}locundefined`)
    })

    it('Returns the correct unique id if version specified', function () {
      const e = new EntryReq({ hash, location, version: 2 })
      expect(e.uniqueId).to.equal(`${hash}loc2`)
    })
  })

  it('Can create a valid entry-req', function () {
    const e = new EntryReq({ hash, location })
    expect(e.hash).to.equal(hash)
    expect(e.location).to.equal(location)
    expect(e.version).to.equal(undefined)
  })

  it('Can create a valid versioned entry-req', function () {
    const e = new EntryReq({ hash, location, version: 5 })
    expect(e.hash).to.equal(hash)
    expect(e.location).to.equal(location)
    expect(e.version).to.equal(5)
  })

  describe('Test invalid entryReqs throw', function () {
    const invalids = [
      [{ hash, location: '' }, 'empty location'],
      [{ hash: 'a'.repeat(63), location }, 'hash too short'],
      [{ hash, location, version: 'Nope' }, 'version no int'],
      [{ hash }, 'No location'],
      [{ location }, 'No hash']

    ]

    invalids.forEach(([entryReq, desc]) => {
      it(`throws on invalid entry req (${desc})`, function () {
        expect(() => new EntryReq(entryReq)).to.throw()
      })
    })
  })

  describe('Test getting the entry', function () {
    let hyperInterface
    let hyperbee, entryReq
    const entry = 'I am an entry'
    const loc = 'someLoc'
    let getEntryFunc

    this.beforeEach(async function () {
      hyperInterface = await hyperInterfaceFactory()

      hyperbee = await hyperInterface.createHyperbee('testbee')
      await hyperbee.put(loc, entry)

      entryReq = new EntryReq(
        { hash: hyperbee.feed.key, location: loc }
      )

      getEntryFunc = async (entry) => await hyperInterface.getEntry(entry)
    })

    this.afterEach(async function () {
      await hyperInterface.close()
    })

    it('Can get the entry', async function () {
      expect((await entryReq.get(getEntryFunc))).to.equal('I am an entry')
    })

    it('Can get an old entry version', async function () {
      const version = hyperbee.version

      const newEntry = 'new entry'
      await hyperbee.put(loc, newEntry)

      const oldEntryReq = new EntryReq(
        { hash: hyperbee.feed.key, location: loc, version }
      )
      expect((await oldEntryReq.get(getEntryFunc))).to.equal(entry)

      // Sanity check
      expect((await entryReq.get(getEntryFunc))).to.equal('new entry')
    })

    it('Does not include the version if undefined for toJson', function () {
      expect(entryReq.toJson()).to.deep.equal(
        { [c.HASH]: utils.getHashAsStr(hyperbee.feed.key), [c.LOCATION]: loc }
      )
    })

    it('Does include the version if defined for toJson', function () {
      entryReq = new EntryReq({ ...entryReq, version: 1 })
      expect(entryReq.toJson()).to.deep.equal(
        {
          [c.HASH]: utils.getHashAsStr(hyperbee.feed.key),
          [c.LOCATION]: loc,
          [c.VERSION]: 1,
          [c.VERSION]: 1
        }
      )
    })
  })
})
