const { expect } = require('chai')
const { strict: nodeAssert } = require('assert')

const utils = require('../lib/utils')
const { hyperInterfaceFactory } = require('./fixtures')
const c = require('../lib/constants')
const EntryReq = require('../lib/entry-request')
const { DiGraph } = require('..')
const BeeInterface = require('../lib/bee-interface')
const Node = require('../lib/node')

describe('Test DiGraph', function () {
  const locRoot = c.ROOT
  const locEntry1 = 'locEntry1'
  const locEntry2 = 'locEntry2'
  const contRoot = 'Content root'
  const contEntry1 = 'Content entry1'
  const contEntry2 = 'Content entry2'

  let hyperInterface, bee
  let rootEntryReq
  let entry1, entry2, rootEntry
  let delayedBackend, backend
  let diGraph
  let child1, child2

  this.beforeEach(async function () {
    hyperInterface = await hyperInterfaceFactory()
    backend = new BeeInterface(hyperInterface)

    const delayedGetEntryFunc = async (entry) => {
      // Simulates peer-finding + network delay
      await new Promise(resolve => setTimeout(resolve, 100))
      return await hyperInterface.getEntry(entry)
    }
    delayedBackend = new BeeInterface(hyperInterface)
    delayedBackend.getEntry = delayedGetEntryFunc

    child1 = { [c.LOCATION]: locEntry1 }
    child2 = { [c.LOCATION]: locEntry2 }
    entry1 = {
      [c.CONTENT]: contEntry1,
      [c.CHILDREN]: []
    }

    entry2 = {
      [c.CONTENT]: contEntry2,
      [c.CHILDREN]: []
    }

    rootEntry = {
      [c.CONTENT]: contRoot,
      [c.CHILDREN]: [child1, child2]
    }

    bee = await hyperInterface.createHyperbee('testbee')
    await bee.put(locEntry1, entry1)
    await bee.put(locEntry2, entry2)
    await bee.put(locRoot, rootEntry)

    rootEntryReq = new EntryReq(
      { hash: bee.feed.key, location: locRoot }
    )

    diGraph = new DiGraph({ rootEntryReq, backend })
  })

  this.afterEach(async function () {
    await hyperInterface.close()
  })

  it('Allows passing custom node class', async function () {
    // Sanity check
    const allNodes = await _getAllNodes(diGraph)
    expect(allNodes[0].extraFunc).to.equal(undefined)

    diGraph = new DiGraph({ rootEntryReq, backend, NodeClass: CustomNode })
    const allCustomNodes = await _getAllNodes(diGraph)
    expect(allCustomNodes[0].extraFunc()).to.equal('I am custom node')
  })

  it('Can create a new digraph with createNew', async function () {
    diGraph = await DiGraph.createNew({ name: 'my graph', hyperInterface })
    const allNodes = await _getAllNodes(diGraph)

    expect(allNodes.length).to.equal(1)
    expect((await allNodes[0].getContent())).to.deep.equal({})
    expect((await allNodes[0].getChildren())).to.deep.equal([])
  })

  it('createNew passes options to constructor', async function () {
    diGraph = await DiGraph.createNew(
      { name: 'my graph', hyperInterface, NodeClass: CustomNode }
    )
    const allNodes = await _getAllNodes(diGraph)

    expect((await allNodes[0].extraFunc())).to.equal('I am custom node')
  })

  it('Can load a digraph from a bee', async function () {
    diGraph = await DiGraph.loadFromBee({ hash: bee.feed.key, hyperInterface })
    const allNodes = await _getAllNodes(diGraph)

    expect(allNodes.length).to.equal(3)
    expect((await allNodes[0].getEntry())).to.deep.equal(rootEntry)
  })

  it('passes options to constructor for loadFromBee', async function () {
    diGraph = await DiGraph.loadFromBee(
      { hash: bee.feed.key, hyperInterface, NodeClass: CustomNode }
    )
    const allNodes = await _getAllNodes(diGraph)

    expect((await allNodes[0].extraFunc())).to.equal('I am custom node')
  })

  it('Throws when loading a digraph from bee without standard root location', async function () {
    await bee.del(locRoot)
    await bee.put('otherLoc', rootEntry)

    await nodeAssert.rejects(
      DiGraph.loadFromBee({ hash: bee.feed.key, hyperInterface }),
      { message: /No entry at 'root'.+/ }
    )
  })

  it('Can get the root hash of a digraph', async function () {
    expect(diGraph.rootHash).to.deep.equal(utils.getHashAsStr(bee.feed.key))
  })

  it('Can walkNoRepeatVisits', async function () {
    const allNodes = await _getAllNodes(diGraph)

    const expectedContents = new Set([contRoot, contEntry1, contEntry2])
    expect(allNodes.length).to.equal(expectedContents.size)

    const actualContents = await _getNodeContentSet(allNodes)
    expect(expectedContents).to.deep.equal(actualContents)
  })

  it('Can walkNoRepeatVisits with loops', async function () {
    entry2[c.CHILDREN].push(child1)

    const allNodes = await _getAllNodes(diGraph)

    const expectedContents = new Set([contRoot, contEntry1, contEntry2])
    expect(allNodes.length).to.equal(expectedContents.size)

    const actualContents = await _getNodeContentSet(allNodes)
    expect(expectedContents).to.deep.equal(actualContents)
  })

  it('Simultaneously awaits all children in walkNoRepeatVisits', async function () {
    const entryGenerator = entryGeneratorFunc(3, bee)

    for (let i = 3; i < 10; i++) {
      const entry = (await entryGenerator.next()).value
      rootEntry[c.CHILDREN].push({ [c.LOCATION]: entry[c.LOCATION] })
    }
    await bee.put(locRoot, rootEntry)

    diGraph = new DiGraph({ rootEntryReq, backend: delayedBackend })

    const startTime = new Date()
    const allNodes = await _getAllNodes(diGraph)
    const timeMs = new Date() - startTime

    // non-simultaneous exec would be > 10*100ms
    // Underbound here is 2*100 (root->entry-x)
    expect(timeMs).to.be.lessThan(500)
    expect(allNodes.length).to.equal(10) // Sanity check
  })

  it('Simultaneous awaits for walkNoRepeatVisits -- complex scenario', async function () {
    const entryGenerator = entryGeneratorFunc(3, bee)

    const entry1Child1 = (await entryGenerator.next()).value
    const entry1Child2 = (await entryGenerator.next()).value
    const entry1Child3 = (await entryGenerator.next()).value
    entry1[c.CHILDREN] = [entry1Child1, entry1Child2, entry1Child3]
    await bee.put(locEntry1, entry1)

    const entry2Child1 = (await entryGenerator.next()).value
    const entry2Child2 = (await entryGenerator.next()).value
    const entry2Child3 = (await entryGenerator.next()).value
    entry2[c.CHILDREN] = [entry2Child1, entry2Child2, entry2Child3]
    await bee.put(locEntry2, entry2)

    const lvl3Child1 = (await entryGenerator.next()).value
    const lvl3Child2 = (await entryGenerator.next()).value
    const lastEntryLoc = entry2Child3[c.LOCATION]
    const { value: lastEntry } = await bee.get(lastEntryLoc)
    lastEntry[c.CHILDREN] = [lvl3Child1, lvl3Child2]
    await bee.put(lastEntryLoc, lastEntry)

    diGraph = new DiGraph({ rootEntryReq, backend: delayedBackend })

    const startTime = new Date()
    const allNodes = await _getAllNodes(diGraph)
    const timeMs = new Date() - startTime

    // non-simultaneous exec would be > 11*100ms
    // Best possible here is 4*100ms (root->entry2->entry2child3->lvl3Child2)
    expect(timeMs).to.be.lessThan(700) // Some margin
    expect(allNodes.length).to.equal(11) // Sanity check
  })

  it('Deep validates correct digraph', async function () {
    await nodeAssert.doesNotReject(diGraph.deepValidate())
  })

  describe('Throws when deep-validating incorrect digraph', async function () {
    it('Throws on invalid child node', async function () {
      delete entry2[c.CONTENT]
      await bee.put(locEntry2, entry2)

      await nodeAssert.rejects(
        diGraph.deepValidate(), { message: /Invalid node entry.+property 'content'.+/ }
      )
    })

    it('Throws on child not found', async function () {
      await bee.del(locEntry1)

      await nodeAssert.rejects(
        diGraph.deepValidate(), { message: /No entry at 'locEntry1' for.+/ }
      )
    })
  })
})

async function _getAllNodes (diGraph) {
  const allNodes = []
  for await (const node of diGraph.walkNoRepeatVisits()) {
    allNodes.push(node)
  }
  return allNodes
}

async function _getNodeContentSet (nodes) {
  const contents = new Set(
    await Promise.all(nodes.map(async (node) => await node.getContent()))
  )
  return contents
}

async function * entryGeneratorFunc (startI, bee) {
  let entryNr = startI
  while (true) {
    const entry = {
      [c.CONTENT]: `Content ${entryNr}`,
      [c.CHILDREN]: []
    }
    const loc = `entry ${entryNr}`
    await bee.put(loc, entry)
    yield { [c.LOCATION]: loc }
    entryNr++
  }
}

class CustomNode extends Node {
  extraFunc () {
    return 'I am custom node'
  }
}
